# Copyright (c) 2012, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Tests for ACPI"""

import acpi
import bits
import struct
import testutil
import testsuite
import time

def register_tests():
    testsuite.add_test("ACPI _MAT (Multiple APIC Table Entry) under Processor objects", test_mat, submenu="ACPI Tests")
    testsuite.add_test("ACPI _PSS (Pstate) table conformance tests", test_pss, submenu="ACPI Tests")
    testsuite.add_test("ACPI _PSS (Pstate) runtime tests", test_pstates, submenu="ACPI Tests")

def test_mat():
    cpupaths = acpi.get_cpupaths()
    procid_apicid, uid_x2apicid = acpi.parse_madt()
    for cpupath in cpupaths:
        # Find the ProcId defined by the processor object
        processor = acpi.evaluate(cpupath)
        # Find the UID defined by the processor object's _UID method
        uid = acpi.evaluate(cpupath + "._UID")
        mat_buffer = acpi.evaluate(cpupath + "._MAT")
        if mat_buffer is None:
            continue
        # Process each _MAT subtable
        i = 0
        while i != len(mat_buffer):
            subtable = acpi.SubTable._make(struct.unpack_from(acpi.SubTable.fmt_str, mat_buffer, i))
            if subtable.Type == acpi.MADT_TYPE_LOCAL_APIC:
                localApic = acpi.MadtLocalApic._make(struct.unpack_from(acpi.MadtLocalApic.fmt_str, mat_buffer, i))
                if localApic.Flags & acpi.MADT_ENABLED:
                    testsuite.test("{} Processor declaration ProcId = _MAT ProcId".format(cpupath), processor.ProcId == localApic.ProcId)
                    testsuite.print_detail("{} ProcId ({:#02x}) != _MAT ProcId ({:#02x})".format(cpupath, processor.ProcId, localApic.ProcId))
                    testsuite.print_detail("Processor Declaration: {}".format(processor))
                    testsuite.print_detail("_MAT offset[{}]: {}".format(i, localApic))
                    if testsuite.test("{} with local APIC in _MAT has local APIC in MADT".format(cpupath), processor.ProcId in procid_apicid):
                        testsuite.test("{} ApicId derived using Processor declaration ProcId = _MAT ApicId".format(cpupath), procid_apicid[processor.ProcId] == localApic.ApicId)
                        testsuite.print_detail("{} ApicId derived from MADT ({:#02x}) != _MAT ApicId ({:#02x})".format(cpupath, procid_apicid[processor.ProcId], localApic.ApicId))
                        testsuite.print_detail("Processor Declaration: {}".format(processor))
                        testsuite.print_detail("_MAT offset[{}]: {}".format(i, localApic))
            if subtable.Type == acpi.MADT_TYPE_LOCAL_X2APIC:
                x2Apic = acpi.MadtLocalx2Apic._make(struct.unpack_from(acpi.MadtLocalx2Apic.fmt_str, mat_buffer, i))
                if x2Apic.Flags & acpi.MADT_ENABLED:
                    if testsuite.test("{} with x2Apic in _MAT has _UID".format(cpupath), uid is not None):
                        testsuite.test("{}._UID = _MAT UID".format(cpupath), uid == x2Apic.UID)
                        testsuite.print_detail("{}._UID ({:#x}) != _MAT UID ({:#x})".format(cpupath, uid, x2Apic.UID))
                        testsuite.print_detail("_MAT offset[{}]: {}".format(i, x2Apic))
                    if testsuite.test("{} with _MAT x2Apic has x2Apic in MADT".format(cpupath), x2Apic.UID in uid_x2apicid):
                        testsuite.test("{} x2ApicId derived from MADT using UID = _MAT x2ApicId".format(cpupath), uid_x2apicid[x2Apic.UID] == x2Apic.X2ApicId)
                        testsuite.print_detail("{} x2ApicId derived from MADT ({:#02x}) != _MAT x2ApicId ({:#02x})".format(cpupath, uid_x2apicid[x2Apic.UID], x2Apic.X2ApicId))
                        testsuite.print_detail("_MAT offset[{}]: {}".format(i, x2Apic))
            i += subtable.Length

def test_pss():
    uniques = acpi.parse_pss()
    for pstates, cpupaths in uniques.iteritems():
        if not testsuite.test("_PSS must exist", pstates is not None):
            testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
            testsuite.print_detail('No _PSS exists')
            continue

        if not testsuite.test("_PSS must not be empty", pstates != ()):
            testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
            testsuite.print_detail('_PSS is empty')
            continue

        testsuite.test("_PSS must contain at most 16 Pstates", len(pstates) <= 16)
        testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
        map(testsuite.print_detail, acpi.show_pss(pstates))

        unique_pstates = set(p for p in pstates)
        testsuite.test("_PSS must have no duplicate Pstates", len(pstates) == len(unique_pstates))
        testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
        map(testsuite.print_detail, acpi.show_pss(pstates))

        frequencies = [p.CoreFrequency for p in pstates]
        testsuite.test("_PSS must list Pstates in descending order of frequency", frequencies == sorted(frequencies, reverse=True))
        testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
        map(testsuite.print_detail, acpi.show_pss(pstates))

        unique_frequencies = set(frequencies)
        testsuite.test("_PSS must have Pstates with no duplicate frequencies", len(frequencies) == len(unique_frequencies))
        testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
        map(testsuite.print_detail, acpi.show_pss(pstates))

        dissipations = [p.Power for p in pstates]
        testsuite.test("_PSS must list Pstates in descending order of power dissipation", dissipations == sorted(dissipations, reverse=True))
        testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
        map(testsuite.print_detail, acpi.show_pss(pstates))

def test_pstates():
    """Execute and verify frequency for each Pstate in the _PSS"""
    old_mwait = {}
    try:
        IA32_PERF_CTL = 0x199

        # Force use of MWAIT C3
        hint = 0x20
        cpus = bits.cpus()
        for apicid in cpus:
            old_mwait[apicid] = bits.get_mwait(apicid)
            bits.set_mwait(apicid, True, hint)

        cpupath_procid = acpi.find_procid()
        cpupath_uid = acpi.find_uid()
        procid_apicid, uid_x2apicid = acpi.parse_madt()
        def cpupath_apicid(cpupath):
            procid = cpupath_procid.get(cpupath, None)
            if procid is not None:
                apicid = procid_apicid.get(procid, None)
                if apicid is not None:
                    return apicid
            uid = cpupath_uid.get(cpupath, None)
            if uid is not None:
                apicid = uid_x2apicid.get(uid, None)
                if apicid is not None:
                    return apicid
            return None

        bclk = testutil.adjust_to_nearest(bits.bclk(), 100.0/12) * 1000000

        uniques = acpi.parse_pss()
        for pstates, cpupaths in uniques.iteritems():
            print "Test duration is ~{} seconds...".format(len(pstates) + 2)

            for n, pstate in enumerate(pstates):
                for cpupath in cpupaths:
                    apicid = cpupath_apicid(cpupath)
                    if apicid is None:
                        print 'Failed to find apicid for cpupath {}'.format(cpupath)
                        continue
                    bits.wrmsr(apicid, IA32_PERF_CTL, pstate.Control)

                turbo = (n == 0 and pstate.CoreFrequency == (pstates[1].CoreFrequency + 1))
                if turbo:
                    # Needs to busywait, not sleep
                    start = time.time()
                    while (time.time() - start < 2):
                        pass

                aperf = bits.cpu_frequency()[1]
                aperf = testutil.adjust_to_nearest(aperf, bclk/2)
                aperf = int(aperf / 1000000)

                if turbo:
                    testsuite.test("P{}: Turbo measured frequency {} >= expected {} MHz".format(n, aperf, pstate.CoreFrequency), aperf >= pstate.CoreFrequency)
                else:
                    testsuite.test("P{}: measured frequency {} MHz == expected {} MHz".format(n, aperf, pstate.CoreFrequency), aperf == pstate.CoreFrequency)
    finally:
        for apicid, old_mwait_values in old_mwait.iteritems():
            bits.set_mwait(apicid, *old_mwait_values)

def test_psd_thread_scope():
    uniques = acpi.parse_psd()
    if not testsuite.test("_PSD (P-State Dependency) must exist for each processor", None not in uniques):
        testsuite.print_detail(acpi.factor_commonprefix(uniques[None]))
        testsuite.print_detail('No _PSD exists')
        return
    unique_NumEntries = {}
    unique_Revision = {}
    unique_Domain = {}
    unique_CoordType = {}
    unique_NumProcessors = {}
    for value, cpupaths in uniques.iteritems():
        unique_NumEntries.setdefault(value.NumEntries, []).extend(cpupaths)
        unique_Revision.setdefault(value.Revision, []).extend(cpupaths)
        unique_Domain.setdefault(value.Domain, []).extend(cpupaths)
        unique_CoordType.setdefault(value.CoordType, []).extend(cpupaths)
        unique_NumProcessors.setdefault(value.NumProcessors, []).extend(cpupaths)

    def detail(d, fmt):
        for value, cpupaths in sorted(d.iteritems(), key=(lambda (k,v): v)):
            testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
            testsuite.print_detail(fmt.format(value))

    testsuite.test('_PSD.NumEntries must be 5', unique_NumEntries.keys() == [5])
    detail(unique_NumEntries, 'Numtries = {} (Expected 5)')
    testsuite.test('_PSD.Revision must be 0', unique_Revision.keys() == [0])
    detail(unique_Revision, 'Revision = {}')
    testsuite.test('_PSD.CoordType must be 0xFE (HW_ALL)', unique_CoordType.keys() == [0xfe])
    detail(unique_CoordType, 'CoordType = {:#x} (Expected 0xFE HW_ALL)')
    testsuite.test('_PSD.Domain must be unique (thread-scoped) for each processor', len(unique_Domain) == len(acpi.get_cpupaths()))
    detail(unique_Domain, 'Domain = {:#x} (Expected a unique value for each processor)')
    testsuite.test('_PSD.NumProcessors must be 1', unique_NumProcessors.keys() == [1])
    detail(unique_NumProcessors, 'NumProcessors = {} (Expected 1)')
